@@ -30,11 +30,14 @@ if TYPE_CHECKING:
 @liquid_filter
 def size(obj: Any) -> int:
     """Return the length of an array or string."""
-    return len(obj)
+    try:
+        return len(obj)
+    except TypeError:
+        return 0
 
 
 @liquid_filter
-def default(obj: Any, default_: object, *, allow_false: bool = False) -> Any:
+def default(obj: Any, default_: object = "", *, allow_false: bool = False) -> Any:
     """Return a default value if the input is nil, false, or empty."""
     _obj = obj
     if hasattr(obj, "__liquid__"):
@@ -43,7 +46,7 @@ def default(obj: Any, default_: object, *, allow_false: bool = False) -> Any:
     if allow_false is True and _obj is False:
         return obj
 
-    if _obj in (None, False, EMPTY) or is_undefined(_obj):
+    if is_undefined(_obj) or _obj in (None, False, EMPTY):
         return default_
 
     return obj
@@ -60,12 +63,10 @@ def date(
         return ""
 
     if is_undefined(fmt):
-        raise FilterArgumentError(
-            f"expected a format string, found {type(fmt).__name__}"
-        )
+        return str(dat)
 
     if isinstance(dat, str):
-        if dat == "now":
+        if dat in ("now", "today"):
             dat = datetime.datetime.now()
         else:
             dat = parser.parse(dat)
