@@ -4,15 +4,23 @@ from __future__ import annotations
 
 import collections.abc
 import datetime
-import functools
 import itertools
 import re
+import sys
 import warnings
 
 from collections import deque
 from contextlib import contextmanager
+
+from functools import lru_cache
+from functools import partial
+from functools import reduce
+
 from itertools import cycle
+from io import StringIO
+
 from operator import getitem
+from operator import mul
 
 from typing import Any
 from typing import Callable
@@ -23,29 +31,30 @@ from typing import List
 from typing import Mapping
 from typing import Optional
 from typing import Sequence
+from typing import TextIO
 from typing import Union
 from typing import TYPE_CHECKING
 
 from liquid.exceptions import NoSuchFilterFunc
 from liquid.exceptions import ContextDepthError
 from liquid.exceptions import Error
+from liquid.exceptions import LocalNamespaceLimitError
+from liquid.exceptions import LoopIterationLimitError
 from liquid.exceptions import UndefinedError
 from liquid.exceptions import lookup_warning
 
 from liquid import Mode
+from liquid.output import LimitedStringIO
 
 if TYPE_CHECKING:  # pragma: no cover
     from liquid import Environment
     from liquid.template import BoundTemplate
     from liquid.builtin.tags.for_tag import ForLoop
 
-# Maximum number of times a context can be extended or wrapped.
-MAX_CONTEXT_DEPTH = 30
 
 ContextPath = Union[str, Sequence[Union[str, int]]]
 Namespace = Mapping[str, object]
 
-
 _undefined = object()
 
 
@@ -143,7 +152,7 @@ def get_item(
 ) -> Any:
     """Chained item getter."""
     try:
-        itm: Any = functools.reduce(_getitem, items, obj)
+        itm: Any = reduce(_getitem, items, obj)
     except (KeyError, IndexError, TypeError):
         itm = default
     return itm
@@ -287,24 +296,25 @@ class StrictUndefined(Undefined):
 
     __slots__ = ("msg",)
 
-    def __init__(self, name: str, obj: object = _undefined, hint: Optional[str] = None):
-        super().__init__(name, obj=obj, hint=hint)
-        if self.hint:
-            self.msg = self.hint
-        else:
-            self.msg = f"'{self.name}' is undefined"
-
-    def __getattribute__(self, name: str) -> Any:
-        if name in (
+    # Properties that don't raise an UndefinedError.
+    allowed_properties = frozenset(
+        [
             "__repr__",
             "name",
             "hint",
             "obj",
-            "hint",
             "msg",
-        ):
-            return super().__getattribute__(name)
-        raise UndefinedError(self.msg)
+        ]
+    )
+
+    def __init__(self, name: str, obj: object = _undefined, hint: Optional[str] = None):
+        super().__init__(name, obj=obj, hint=hint)
+        self.msg = self.hint if self.hint else f"'{self.name}' is undefined"
+
+    def __getattribute__(self, name: str) -> object:
+        if name in object.__getattribute__(self, "allowed_properties"):
+            return object.__getattribute__(self, name)
+        raise UndefinedError(object.__getattribute__(self, "msg"))
 
     def __contains__(self, item: object) -> bool:
         raise UndefinedError(self.msg)
@@ -340,7 +350,24 @@ class StrictUndefined(Undefined):
         raise UndefinedError(self.msg)
 
 
-# pylint: disable=too-many-instance-attributes redefined-builtin
+class StrictDefaultUndefined(StrictUndefined):
+    """An undefined that plays nicely with the `default` filter."""
+
+    # Properties that don't raise an UndefinedError.
+    allowed_properties = frozenset(
+        [
+            "__repr__",
+            "__liquid__",
+            "__class__",
+            "name",
+            "hint",
+            "obj",
+            "msg",
+        ]
+    )
+
+
+# pylint: disable=too-many-instance-attributes redefined-builtin too-many-public-methods
 class Context:
     """Liquid template context."""
 
@@ -356,11 +383,14 @@ class Context:
         "autoescape",
         "_copy_depth",
         "parent_context",
+        "loop_iteration_carry",
+        "local_namespace_size_carry",
     )
 
     getitem = _getitem
     getitem_async = _getitem_async
 
+    # pylint: disable=too-many-arguments
     def __init__(
         self,
         env: Environment,
@@ -368,6 +398,8 @@ class Context:
         disabled_tags: Optional[List[str]] = None,
         copy_depth: int = 0,
         parent_context: Optional[Context] = None,
+        loop_iteration_carry: int = 1,
+        local_namespace_size_carry: int = 0,
     ):
         self.env = env
 
@@ -395,7 +427,7 @@ class Context:
         }
 
         # As stack of forloop objects. Used for populating forloop.parentloop.
-        self.loops: List[object] = []
+        self.loops: List[ForLoop] = []
 
         # A list of tags names that are disallowed in this context. For example,
         # partial templates rendered using the "render" tag are not allowed to
@@ -413,9 +445,31 @@ class Context:
         # A reference to a parent render context, if this one has been copied.
         self.parent_context = parent_context
 
+        # A loop iteration count carried over from a parent context, if this one has
+        # been copied. This helps us adhere to loop iteration limits in templates
+        # rendered with the `render` tag.
+        self.loop_iteration_carry = loop_iteration_carry
+
+        # The cumulative number of names in parent render context local namespaces. We
+        # use this to adhere to local namespace limits when copying a render context.
+        self.local_namespace_size_carry = local_namespace_size_carry
+
     def assign(self, key: str, val: Any) -> None:
         """Add `val` to the context with key `key`."""
         self.locals[key] = val
+        if (
+            self.env.local_namespace_limit
+            and self._get_size_of_locals() > self.env.local_namespace_limit
+        ):
+            raise LocalNamespaceLimitError("local namespace limit reached")
+
+    def _get_size_of_locals(self) -> int:
+        if not self.env.local_namespace_limit:
+            return 0
+        return (
+            sum(sys.getsizeof(obj, default=1) for obj in set(self.locals.values()))
+            + self.local_namespace_size_carry
+        )
 
     def get(self, path: ContextPath, default: object = _undefined) -> object:
         """Return the value at path `path` if it is in scope, else default."""
@@ -428,7 +482,7 @@ class Context:
 
         if items:
             try:
-                return functools.reduce(Context.getitem, items, obj)
+                return reduce(Context.getitem, items, obj)
             except (KeyError, IndexError, TypeError) as err:
                 if isinstance(err, KeyError):
                     hint = (
@@ -487,7 +541,7 @@ class Context:
                 return self.env.undefined(name)
             return default
 
-    @functools.lru_cache(maxsize=128)
+    @lru_cache(maxsize=128)
     def filter(self, name: str) -> Callable[..., object]:
         """Return the filter function with given name."""
         try:
@@ -504,7 +558,7 @@ class Context:
             kwargs["environment"] = self.env
 
         if kwargs:
-            return functools.partial(filter_func, **kwargs)
+            return partial(filter_func, **kwargs)
 
         return filter_func
 
@@ -571,7 +625,7 @@ class Context:
     @contextmanager
     def extend(self, namespace: Namespace) -> Iterator[Context]:
         """Extend this context with the given read-only namespace."""
-        if self.scope.size() > MAX_CONTEXT_DEPTH:
+        if self.scope.size() > self.env.context_depth_limit:
             raise ContextDepthError(
                 "maximum context depth reached, possible recursive include"
             )
@@ -586,6 +640,7 @@ class Context:
     @contextmanager
     def loop(self, namespace: Namespace, forloop: ForLoop) -> Iterator[Context]:
         """Just like ``Context.extend``, but keeps track of ForLoop objects too."""
+        self.raise_for_loop_limit(forloop.length)
         self.loops.append(forloop)
         with self.extend(namespace) as context:
             try:
@@ -600,22 +655,51 @@ class Context:
         except IndexError:
             return self.env.undefined("parentloop")
 
+    def raise_for_loop_limit(self, length: int = 1) -> None:
+        """Raise a ``LoopIterationLimitError`` if the product of the loop stack is
+        greater than the configured loop iteration limit.
+        """
+        if (
+            self.env.loop_iteration_limit
+            and reduce(
+                mul,
+                itertools.chain(
+                    (loop.length for loop in self.loops),
+                    [length, self.loop_iteration_carry],
+                ),
+                1,
+            )
+            > self.env.loop_iteration_limit
+        ):
+            raise LoopIterationLimitError("loop iteration limit reached")
+
     def copy(
-        self, namespace: Namespace, disabled_tags: Optional[List[str]] = None
+        self,
+        namespace: Namespace,
+        disabled_tags: Optional[List[str]] = None,
+        carry_loop_iterations: bool = False,
     ) -> Context:
         """Return a copy of this context without any local variables or other state
         for stateful tags."""
-        if self._copy_depth > MAX_CONTEXT_DEPTH:
+        if self._copy_depth > self.env.context_depth_limit:
             raise ContextDepthError(
                 "maximum context depth reached, possible recursive render"
             )
 
-        return type(self)(
+        loop_iteration_carry = (
+            reduce(mul, (loop.length for loop in self.loops), 1)
+            if carry_loop_iterations
+            else 1
+        )
+
+        return self.__class__(
             self.env,
             globals=ReadOnlyChainMap(namespace, self.globals),
             disabled_tags=disabled_tags,
             copy_depth=self._copy_depth + 1,
             parent_context=self,
+            loop_iteration_carry=loop_iteration_carry,
+            local_namespace_size_carry=self._get_size_of_locals(),
         )
 
     def error(self, exc: Error) -> None:
@@ -625,13 +709,30 @@ class Context:
         if self.env.mode == Mode.WARN:
             warnings.warn(str(exc), category=lookup_warning(exc.__class__))
 
+    def get_buffer(self, buf: Optional[TextIO] = None) -> StringIO:
+        """Return a new StringIO object, possibly limited according to the configured
+        output stream limit."""
+        if self.env.output_stream_limit is None:
+            return StringIO()
+
+        carry = buf.size if isinstance(buf, LimitedStringIO) else 0
+        return LimitedStringIO(limit=self.env.output_stream_limit - carry)
+
 
 class VariableCaptureContext(Context):
     """A render context that captures template variable names."""
 
+    __slots__ = (
+        "local_references",
+        "all_references",
+        "undefined_references",
+        "root_context",
+    )
+
     # Used for formatting context path strings.
     re_ident = re.compile(r"^[\w_][\w_\-]*$")
 
+    # pylint: disable=too-many-arguments
     def __init__(
         self,
         env: Environment,
@@ -639,8 +740,18 @@ class VariableCaptureContext(Context):
         disabled_tags: Optional[List[str]] = None,
         copy_depth: int = 0,
         parent_context: Optional[VariableCaptureContext] = None,
+        loop_iteration_carry: int = 1,
+        local_namespace_size_carry: int = 0,
     ):
-        super().__init__(env, globals, disabled_tags, copy_depth, parent_context)
+        super().__init__(
+            env,
+            globals=globals,
+            disabled_tags=disabled_tags,
+            copy_depth=copy_depth,
+            parent_context=parent_context,
+            loop_iteration_carry=loop_iteration_carry,
+            local_namespace_size_carry=local_namespace_size_carry,
+        )
         self.local_references: List[str] = []
         self.all_references: List[str] = []
         self.undefined_references: List[str] = []
